from frasco.ext import *
from flask_assets import Environment, _webassets_cmd
from flask import Blueprint
from flask.cli import with_appcontext, cli
from flask.signals import Namespace as SignalNamespace
import logging
import click


_signals = SignalNamespace()
before_build_assets = _signals.signal('before_build_assets')
after_clean_assets = _signals.signal('before_clean_assets')
auto_build_assets = _signals.signal('auto_build_assets')


class FrascoAssetsState(ExtensionState):
    def register(self, *args, **kwargs):
        return self.env.register(*args, **kwargs)


class FrascoAssets(Extension):
    name = 'frasco_assets'
    state_class = FrascoAssetsState
    prefix_extra_options = 'ASSETS_'

    def _init_app(self, app, state):
        if app.debug:
            app.config.setdefault('ASSETS_AUTO_BUILD', True)
        state.env = Environment(app)
        state.env.debug = app.debug

        @app.cli.command()
        @with_appcontext
        def build_all_assets():
            """Build assets from all extensions."""
            before_build_assets.send()
            _webassets_cmd('build')

        if state.env.config["auto_build"]:
            @app.before_first_request
            def before_request():
                auto_build_assets.send(self)

    @ext_stateful_method
    def register(self, state, *args, **kwargs):
        return state.register(*args, **kwargs)


class AssetsBlueprint(Blueprint):
    def __init__(self, name, import_name, **kwargs):
        kwargs.setdefault('static_url_path', '/static/vendor/%s' % name)
        kwargs.setdefault('static_folder', 'static')
        super(AssetsBlueprint, self).__init__(name, import_name, **kwargs)


def expose_package(app, name, import_name):
    app.register_blueprint(AssetsBlueprint(name, import_name))


def register_assets_builder(func=None):
    def decorator(func):
        before_build_assets.connect(lambda sender: func(), weak=False)
        auto_build_assets.connect(lambda sender: func(), weak=False)
    if func:
        return decorator(func)
    return decorator



from flask import has_app_context, current_app, has_request_context, session, request
from frasco.users import user_login_context, is_user_logged_in, current_user
from frasco.utils import import_string
from flask_rq2.job import FlaskJob
from rq.job import unpickle, UNEVALUATED, dumps, Job as RQJob


def pack_task_args(data):
    """Traverse data and converts every object with a __taskdump__() method
    """
    if hasattr(data, "__taskdump__"):
        cls, state = data.__taskdump__()
        if not cls:
            cls = data.__class__.__module__ + "." + data.__class__.__name__
        return {"$taskobj": [cls, state]}
    if isinstance(data, (list, tuple)):
        lst = []
        for item in data:
            lst.append(pack_task_args(item))
        return lst
    if isinstance(data, dict):
        dct = {}
        for k, v in data.iteritems():
            dct[k] = pack_task_args(v)
        return dct
    return data


def unpack_task_args(data):
    """Traverse data and transforms back objects which where dumped
    using __taskdump()
    """
    if isinstance(data, (list, tuple)):
        lst = []
        for item in data:
            lst.append(unpack_task_args(item))
        return lst
    if isinstance(data, dict):
        if "$taskobj" in data:
            cls = import_string(data["$taskobj"][0])
            return cls.__taskload__(data["$taskobj"][1])
        else:
            dct = {}
            for k, v in data.iteritems():
                dct[k] = unpack_task_args(v)
            return dct
    return data


class FrascoJob(FlaskJob):
    @classmethod
    def create(cls, func, *args, **kwargs):
        job = super(FrascoJob, cls).create(func, *args, **kwargs)
        if is_user_logged_in():
            job.meta['current_user_id'] = current_user.id
        if has_request_context():
            job.meta.update({
                'endpoint': request.endpoint,
                'remote_addr': request.remote_addr,
                'session': dict(session)
            })
        return job

    def _unpickle_data(self):
        self._func_name, self._instance, self._args, self._kwargs = unpickle(self.data)
        self._args = unpack_task_args(self._args)
        self._kwargs = unpack_task_args(self._kwargs)

    @property
    def data(self):
        if self._data is UNEVALUATED:
            if self._func_name is UNEVALUATED:
                raise ValueError('Cannot build the job data')

            if self._instance is UNEVALUATED:
                self._instance = None

            if self._args is UNEVALUATED:
                args = ()
            else:
                args = pack_task_args(self._args)

            if self._kwargs is UNEVALUATED:
                kwargs = {}
            else:
                kwargs = pack_task_args(self._kwargs)

            job_tuple = self._func_name, self._instance, args, kwargs
            self._data = dumps(job_tuple)
        return self._data

    @data.setter
    def data(self, value):
        self._data = value
        self._func_name = UNEVALUATED
        self._instance = UNEVALUATED
        self._args = UNEVALUATED
        self._kwargs = UNEVALUATED

    def perform(self):
        if has_app_context():
            # if async=False, app context may already exist
            return self.perform_in_app_context()
        app = self.load_app()
        with app.app_context():
            return self.perform_in_app_context()

    def perform_in_app_context(self):
        current_user_id = self.meta.get('current_user_id')
        if current_user_id and not is_user_logged_in(): # user is already logged in if task is async=False
            with user_login_context(current_app.extensions.frasco_users.Model.query.get(current_user_id)):
                return RQJob.perform(self)
        else:
            return RQJob.perform(self)



from flask import request, flash, redirect, session, current_app, Blueprint
from frasco.helpers import url_for
from frasco.ext import get_extension_state
from frasco.users.user import login_user, is_user_logged_in, current_user
from frasco.models import db
from frasco.utils import populate_obj
from frasco.tasks import get_current_job
from frasco.redis.objects import JSONRedisHash
from authlib.flask.client import OAuth
from sqlalchemy.dialects import postgresql
import datetime
import uuid
import hashlib


class OAuth1RequestTokenCache(object):
    def __init__(self):
        self._data = {}

    def get(self, k):
        return self._data.get(k)

    def set(self, k, v, timeout=None):
        self._data[k] = v

    def delete(self, k):
        if k in self._data:
            del self._data[k]


class OAuth1RequestTokenRedisCache(OAuth1RequestTokenCache):
    def __init__(self, *args, **kwargs):
        self._data = JSONRedisHash(*args, **kwargs)


class UserOAuthTokenModelMixin(object):
    created_at = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    provider = db.Column(db.String)
    token_type = db.Column(db.String)
    access_token = db.Column(db.String)
    access_token_secret = db.Column(db.String)
    refresh_token = db.Column(db.String)
    token_expires_at = db.Column(db.DateTime)
    scope = db.Column(db.String)
    data = db.Column(postgresql.JSONB)

    def to_oauth1_token(self):
        return dict(
            oauth_token=self.access_token,
            oauth_token_secret=self.access_token_secret,
        )

    def to_oauth2_token(self):
        o = dict(
            access_token=self.access_token,
            token_type=self.token_type,
            refresh_token=self.refresh_token,
        )
        if self.token_expires_at:
            o['expires_at'] = self.token_expires_at
        return o


class OAuthUserModelMixin(object):
    @classmethod
    def query_by_oauth_token(cls, provider, id_property, id_value):
        return cls.query.join(cls.__oauth_token_model__).filter(
            cls.__oauth_token_model__.data[id_property].astext == unicode(id_value)).first()

    def get_oauth_token(self, provider):
        return self.oauth_tokens.filter(self.__oauth_token_model__.provider == provider).first()

    def save_oauth_token_data(self, provider, data):
        oauth_token = self.get_oauth_token(provider)
        props = {
            'token_type': data.pop('token_type', None),
            'access_token': data.pop('access_token', None),
            'refresh_token': data.pop('refresh_token', None),
            'token_expires_at': data.pop('expires_at', None),
            'scope': data.pop('scope', None),
            'data': data
        }
        if not oauth_token:
            self.oauth_tokens.append(self.__oauth_token_model__(provider=provider, **props))
        else:
            populate_obj(oauth_token, props)

    def remove_oauth_token(self, provider):
        token = self.get_oauth_token(provider)
        if token:
            self.oauth_tokens.remove(token)


oauth = OAuth()


def create_token_user_fetcher(name):
    def fetch_token():
        if is_user_logged_in():
            return current_user.get_oauth_token(name).to_oauth2_token()
    return fetch_token


def create_token_session_fetcher(name):
    def fetch_token():
        job = get_current_job()
        if job:
            return job.meta['session']['%s_oauth_token' % name]
        return session['%s_oauth_token' % name]
    return fetch_token


def create_login_blueprint(name, authorize_handler):
    login_blueprint = Blueprint('%s_login' % name, __name__)

    @login_blueprint.route('/login/%s' % name)
    def login():
        session['oauth_redirect_next'] = request.args.get('next')
        callback_url = url_for('.callback', _external=True)
        return getattr(oauth, name).authorize_redirect(callback_url)

    @login_blueprint.route('/login/%s/callback' % name)
    def callback():
        redirect_url = session.pop('oauth_redirect_next', None)
        token = getattr(oauth, name).authorize_access_token()
        if token is None:
            flash(current_app.extensions.frasco_users.options["oauth_user_denied_login"], "error")
            return redirect(url_for("users.login"))
        try:
            return authorize_handler(token, redirect_url)
        except Exception as e:
            current_app.log_exception(e)
            flash(current_app.extensions.frasco_users.options["oauth_error"], "error")
            return redirect(url_for("users.login"))

    return login_blueprint


def create_session_login_blueprint(name):
    def authorize_handler(token, redirect_url):
        return oauth_session_login(name, token, redirect_url=redirect_url)
    return create_login_blueprint(name, authorize_handler)


def oauth_session_login(provider, token, redirect_url=None):
    state = get_extension_state('frasco_users')
    session['%s_oauth_token' % provider] = token
    if not redirect_url:
        redirect_url = request.args.get('next') or url_for(state.options["redirect_after_login"])
    return redirect(redirect_url)


def oauth_login(provider, id_property, id_value, data, defaults, redirect_url=None):
    """Execute a login via oauth. If no user exists, oauth_signup() will be called
    """
    state = get_extension_state('frasco_users')
    user = state.Model.query_by_oauth_token(provider, id_property, id_value)
    if not redirect_url:
        redirect_url = request.args.get('next') or url_for(state.options["redirect_after_login"])
    if is_user_logged_in():
        if user and user != current_user:
            if state.options["oauth_user_already_exists_message"]:
                flash(state.options["oauth_user_already_exists_message"].format(provider=provider), "error")
            return redirect(redirect_url)
        user = current_user
        if provider not in user.auth_providers:
            user.auth_providers.append(provider)
    elif not user:
        return oauth_signup(provider, data, defaults, redirect_url=redirect_url)
    else:
        login_user(user, provider=provider)
    user.save_oauth_token_data(provider, data)
    return redirect(redirect_url)


def oauth_signup(provider, data, defaults, redirect_url=None):
    """Start the signup process after having logged in via oauth
    """
    session["oauth_signup"] = provider
    session["oauth_data"] = data
    session["oauth_user_defaults"] = defaults
    if not redirect_url:
        redirect_url = request.args.get("next")
    return redirect(url_for('users.oauth_signup', next=redirect_url))


def clear_oauth_signup_session():
    session.pop("oauth_signup", None)
    session.pop("oauth_data", None)
    session.pop("oauth_user_defaults", None)



from frasco import current_app
from frasco.ctx import ContextStack
from contextlib import contextmanager
import logging

bulk_connection_context = ContextStack()


class BulkConnection(object):
    def __init__(self, sender):
        self.sender = sender

    def send(self, msg):
        self.sender(msg)


class MailProvider(object):
    def __init__(self, app, state, options):
        self.app = app
        self.state = state
        self.options = options

    def send(self, msg):
        raise NotImplementedError()

    def start_bulk_connection(self):
        return BulkConnection(self.send)

    def stop_bulk_connection(self, bulk_connection):
        pass

    @contextmanager
    def bulk_connection(self):
        failed_establish = False
        try:
            conn = self.start_bulk_connection()
        except Exception as e:
            if not self.state.options['silent_failures']:
                raise e
            current_app.log_exception(e)
            logging.getLogger('frasco.mail').warning("Failed establishing bulk connection")
            conn = BulkConnection(self.send)
            failed_establish = True

        bulk_connection_context.push(conn)
        try:
            yield conn
        finally:
            if not failed_establish:
                self.stop_bulk_connection(conn)
            bulk_connection_context.pop()



from flask import json, current_app
from frasco.ext import get_extension_state
from frasco.utils import unknown_value
from .utils import build_object_key
import inspect


__all__ = ('redis_cached_property', 'redis_cached_property_as_json', 'redis_cached_method', 'redis_cached_method_as_json')


class RedisCachedAttribute(object):
    def __init__(self, func, redis=None, key=None, ttl=None, coerce=None,\
                 serializer=None, name=None):
        self.func = func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__
        self._redis = redis
        self.key = key
        self.ttl = ttl
        self.coerce = coerce
        self.serializer = serializer
        self.name = name or self.__name__
        self.cached_property_name = self.__name__ + '_cached'
        self.cache_disabled = False
        self.cache_ignore_current = False
        self.cache_current_ttl = None

    @property
    def redis(self):
        return self._redis or get_extension_state('frasco_redis').connection

    def _set_cached_value(self, key, value, default_ttl=None):
        if self.serializer:
            value = self.serializer.dumps(value)
        ttl = self.cache_current_ttl
        if ttl is None:
            ttl = default_ttl
        if value is None:
            value = 'None'
        if ttl is not None:
            self.redis.setex(key, ttl, value)
        else:
            self.redis.set(key, value)

    def _get_cached_value(self, key):
        if not self.redis.exists(key):
            return unknown_value
        value = self.redis.get(key)
        if value is None or value == 'None':
            return None
        if self.serializer:
            value = self.serializer.loads(value)
        if self.coerce:
            value = self.coerce(value)
        return value

    def _call_func(self, obj, *args, **kwargs):
        self.cache_ignore_current = False
        self.cache_current_ttl = self.ttl
        return self.func(obj, *args, **kwargs)


class RedisCachedProperty(RedisCachedAttribute):
    def __init__(self, func, fset=None, fdel=None, **kwargs):
        super(RedisCachedProperty, self).__init__(func, **kwargs)
        self.fset = fset
        self.fdel = fdel

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__.get(self.cached_property_name, unknown_value)
        if value is unknown_value:
            key = None
            if not self.cache_disabled:
                try:
                    key = self.build_key(obj)
                    value = self._get_cached_value(key)
                except Exception as e:
                    current_app.log_exception(e)
                    value = unknown_value
            if value is unknown_value:
                value = self.get_fresh(obj)
                if not self.cache_disabled and not self.cache_ignore_current and key:
                    self._set_cached_value(key, value,
                        getattr(obj, '__redis_cache_ttl__', None))
            obj.__dict__[self.cached_property_name] = value
        return value

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        self.fset(obj, value)
        self.invalidate(obj)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")
        self.fdel(obj)
        self.invalidate(obj)

    def build_key(self, obj):
        return build_object_key(obj, self.name, self.key)

    def get_cached(self, obj):
        try:
            key = self.build_key(obj)
        except Exception as e:
            current_app.log_exception(e)
            return unknown_value
        return self._get_cached_value(key)

    def get_fresh(self, obj):
        return self._call_func(obj)

    def require_fresh(self, obj):
        obj.__dict__.pop(self.cached_property_name, None)

    def invalidate(self, obj):
        try:
            key = self.build_key(obj)
        except Exception as e:
            current_app.log_exception(e)
            return
        self.redis.delete(key)

    def setter(self, fset):
        self.fset = fset
        return self

    def deleter(self, fdel):
        self.fdel = fdel
        return self


def redis_cached_property(fget=None, **kwargs):
    def decorator(f):
        return RedisCachedProperty(f, **kwargs)
    if fget:
        return decorator(fget)
    return decorator


def redis_cached_property_as_json(fget=None, **kwargs):
    kwargs['serializer'] = json
    return redis_cached_property(fget, **kwargs)


class RedisCachedMethod(RedisCachedAttribute):
    def __get__(self, obj, cls=None):
        self.obj = obj
        return self

    def __call__(self, *args, **kwargs):
        obj = kwargs.pop('__obj__', self.obj)
        value = unknown_value
        if not self.cache_disabled:
            key = None
            try:
                key = self.build_key(args, kwargs, obj)
                value = self._get_cached_value(key)
            except Exception as e:
                current_app.log_exception(e)
                value = unknown_value
        if value is unknown_value:
            value = self._call_func(obj, *args, **kwargs)
            if not self.cache_disabled and not self.cache_ignore_current and key:
                self._set_cached_value(key, value,
                    getattr(self.obj, '__redis_cache_ttl__', None))
        return value

    def cached(self, *args, **kwargs):
        obj = kwargs.pop('__obj__', self.obj)
        try:
            key = self.build_key(args, kwargs, obj)
        except Exception as e:
            current_app.log_exception(e)
            return unknown_value
        return self._get_cached_value(key)

    def fresh(self, *args, **kwargs):
        obj = kwargs.pop('__obj__', self.obj)
        return self._call_func(obj, *args, **kwargs)

    def invalidate(self, *args, **kwargs):
        obj = kwargs.pop('__obj__', self.obj)
        try:
            key = self.build_key(args, kwargs, obj)
        except Exception as e:
            current_app.log_exception(e)
            return
        self.redis.delete(key)

    def build_key(self, args=None, kwargs=None, obj=None):
        if not obj:
            obj = self.obj
        if not args:
            args = []
        if not kwargs:
            kwargs = {}
        at_values = inspect.getcallargs(self.func, obj, *args, **kwargs)
        return build_object_key(obj, self.name, self.key, at_values)


def redis_cached_method(func=None, **kwargs):
    def decorator(f):
        return RedisCachedMethod(f, **kwargs)
    if func:
        return decorator(func)
    return decorator


def redis_cached_method_as_json(func=None, **kwargs):
    kwargs['serializer'] = json
    return redis_cached_method(func, **kwargs)
